// X10 CM11a Macro I/O module

/* TODO:

    Sunrise tolerance
    Read output files with Eprom addresses for downloads

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cm11.h"

//#define DEBUG 1

// The macro def headers
static char gKeyMacroData[] = "MACRODATA";
static char gKeyMacro[] = "MACRO";
static char gKeyTimer[] = "TIMER";
static char gKeyExport[] = "EXPORT";
static char gKeyAlias[] = "ALIAS";
static char gKeySymbol[] = "SYMBOL";
static char gKeyExtModule[] = "EXTDIM";


// Macro Body
static char gEvent[] = "EVENT";
static char gDelay[] = "DELAY";
static char gCall[] = "CALL";
static char gName[] = "NAME";
static char gEndM[] = "ENDM";

// Timer stuff
static char gTStart[] = "START";
static char gTStop[] = "STOP";
static char gTBegin[] = "BEGIN";
static char gTEnd[] = "END";
static char gEndT[] = "ENDT";

// EVENT types
static char gFnAllOff[] = "ALLOFF";
static char gFnAllLightsOn[] = "ALLLON";
static char gFnOn[] = "ON";
static char gFnOff[] = "OFF";
static char gFnDim[] = "DIM";
static char gFnDimTo[] = "DIMTO";
static char gFnBright[] = "BRIGHT";
static char gFnAllLightsOff[] = "ALLLOFF";
static char gFnExtCode[] = "EXTC";
static char gFnHailReq[] = "HAILREQ";
static char gFnHailAck[] = "HAILACK";
static char gFnPresetDim[] = "PREDIM";
static char gFnExtData[] = "EXTD";
static char gFnStatOn[] = "STATON";
static char gFnStatOff[] = "STATOFF";
static char gFnStatReq[] = "STATREQ";

#define MAX_SYMBOLS 128

struct tagSymbol 
{
    unsigned long address;
    char name[32];
} gSymbol[MAX_SYMBOLS];

static int symbolMax = 0;

struct tagMacroData 
{
    int referenceCount;
    char label[32];
    
    struct tagMacroData *next;
    unsigned  epromAddress;

    int       delayMinutes;
    struct tagEvent *firstEvent;
} *gMacroDataHead = NULL;


struct tagMacroInitiator 
{
    struct tagMacroInitiator *next;
    unsigned epromAddress;

    X10_UNIT initiatorUnit;
    int isOnFunction;

    struct tagMacroData *macroData;
} *gMacroInitiatorHead = NULL;

struct tagTimerInitiator 
{
    struct tagTimerInitiator *next;
    unsigned epromAddress;
    
    int dayMask;
    int beginDay;
    int endDay;
    int startMinutes;
    int stopMinutes;

    struct tagMacroData *startMacroData;
    struct tagMacroData *stopMacroData;
} *gTimerInitiatorHead = NULL;

static char monthArray[][5] = {"jan-", "feb-", "mar-", "apr-", "may-", "jun-",
                              "jul-", "aug-", "sep-", "oct-", "nov-", "dec-"
};

static int monthMap[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

// Bitmap array of advanced X10 modules
X10_UNIT_BITMAP gExtModules[16] = {0, 0, 0, 0, 0, 0, 0, 0,
                                   0, 0, 0, 0, 0, 0, 0, 0 };

static void AllocFailed()
{
    fprintf(stderr, "Memory allocation failed\n");
}

static void UnexpectedEnd()
{
    fprintf(stderr,"Unexpected end of stream\n");
}

static int ReadDate(STREAM inStream) 
{
    int i;
    int date = 0;
    
    char token[32];
    if (!GetNextToken(inStream, token, sizeof(token))) {
        UnexpectedEnd();
        return -1;
    }
    
    for (i = 0; i < 12 && strncasecmp(monthArray[i], token, 4); i++) {
        date += monthMap[i];
    }
        
    if (i == 12) {
        StreamError(inStream);
        fprintf(stderr, "Invalid date %s\n", token);
        return -1;
    }
    date += atoi(token+4);
    if (DEBUG_MACRO)
        fprintf(stderr,"Read Date = %d\n", date);

    return date;
}

static int ReadTime(STREAM inStream)
{
    int minutes = 0;
    char *colon;
    
    char token[32];
    if (!GetNextToken(inStream, token, sizeof(token))) {
        UnexpectedEnd();
        return -1;
    }
    
    colon=strchr(token, ':');

    if (colon == NULL) {
        StreamError(inStream);
        fprintf(stderr, "Invalid time %s\n", token);
        return -1;
    }
    *colon++ = '\0';

    minutes = atoi(token)*60 + atoi(colon);
    if (DEBUG_MACRO)
        fprintf(stderr,"Read Time = %d\n", minutes);

    return minutes;
}

        

struct tagEvent *ReadEvent(STREAM inStream)
{
    char token[32];
    char *param;
    struct tagEvent *event;
    X10_UNIT houseCode;
    char *emsg;
    
    if (!GetNextToken(inStream, token, sizeof(token))) {
        UnexpectedEnd();
        return NULL;
    }
    if ((event = calloc(1,sizeof(struct tagEvent))) == NULL) {
        StreamError(inStream);
        AllocFailed();
        return NULL;
    }

    // Separate token an optional parameter
    param=strchr(token,':');
    if (param) {
        *param++ = '\0';
    }
    if (DEBUG_MACRO)
        fprintf(stderr,"ReadEvent = %s/%s\n", token, param?param:"");
    
    emsg = (char *)0;
    if (!strcmp(token, gFnAllOff)) {
        event->function = FN_ALL_UNITS_OFF;
    } else if (!strcmp(token, gFnAllLightsOn)) {
        event->function = FN_ALL_LIGHTS_ON;
    } else if (!strcmp(token, gFnOn)) {
        event->function = FN_ON;
    } else if (!strcmp(token, gFnOff)) {
        event->function = FN_OFF;
    } else if (!strcmp(token, gFnDim)) {
        event->function = FN_DIM;
        event->dimLevel = param?atoi(param):6;	/* "DIM" w/o % means 6% */
    } else if (!strcmp(token, gFnDimTo)) {
        event->function = FN_DIM;
        event->dimAbsolute = 1;
		if (param)
			event->dimLevel = atoi(param);
		else
			emsg = "No parameter for DIMTO";
    } else if (!strcmp(token, gFnBright)) {
        event->function = FN_BRIGHT;
        event->dimLevel = param?atoi(param):6; /* default "BRIGHT" is 6% */
    } else if (!strcmp(token, gFnAllLightsOff)) {
        event->function = FN_ALL_LIGHTS_OFF;
    } else if (!strcmp(token, gFnExtCode)) {
        event->function = FN_EXT_CODE;
		if (param)
			Hex2Bytes(param, event->extended, sizeof(event->extended));
		else
			emsg = "No parameter for EXTC";
    } else if (!strcmp(token, gFnHailReq)) {
        event->function = FN_HAIL_REQUEST;
    } else if (!strcmp(token, gFnHailAck)) {
        event->function = FN_HAIL_ACK;
    } else if (!strcmp(token, gFnPresetDim)) {
        event->function = FN_PRESET_DIM_0;
		if (param)
			event->dimLevel = atoi(param);
		else
			emsg = "No parameter for PREDIM";
    } else if (!strcmp(token, gFnExtData)) {
        event->function = FN_EXT_DATA_XFER;
		if (param)
			Hex2Bytes(param, event->extended, sizeof(event->extended));
		else
			emsg = "No param for EXTD";
    } else if (!strcmp(token, gFnStatOn)) {
        event->function = FN_STATUS_ON;
    } else if (!strcmp(token, gFnStatOff)) {
        event->function = FN_STATUS_OFF;
    } else if (!strcmp(token, gFnStatReq)) {
        event->function = FN_STATUS_REQUEST;
    } else {
		emsg = "Invalid event";
    }
    if (emsg) {
        StreamError(inStream);
        fprintf(stderr,"%s: %s\n", emsg, token);
        free(event);
        return NULL;
    }
    if (DEBUG_MACRO)
        fprintf(stderr," Function decoded as = %x\n", event->function);

    // OK, now we get the housecode and devices
    if (!GetNextToken(inStream, token, sizeof(token))) {
        UnexpectedEnd();
        free(event);
        return NULL;
    }

    if (StrToUnits(token, &houseCode, &event->deviceMask) < 0) {
        StreamError(inStream);
        fprintf(stderr, "Bad Alias\n");
        free(event);
        return NULL;
    }

    if (event->function == FN_PRESET_DIM_0) {
        // Preset dim says to set the dim level in the housecode fields
        // plus MSB in bit 0 of the function field
        event->function |= ((event->dimLevel & 0x0f) << 4) |
            ((event->dimLevel & 0x10) ? 1 : 0);
    } else {
        // Normal events use the housecode
        event->function |= (houseCode & 0xf0);
    }
    if (DEBUG_MACRO)
        fprintf(stderr," Event fn = %x units = %x\n", event->function,
                event->deviceMask);

    return event;
}


static struct tagMacroInitiator
*InsertNewMacroInitiator(struct tagMacroInitiator **head)
{
    struct tagMacroInitiator *new =calloc(1, sizeof(struct tagMacroInitiator));

    if (new) {
        if (*head) {
            struct tagMacroInitiator *tmp = *head;
            while (tmp->next) tmp = tmp->next;
            tmp->next = new;
        }
        else *head = new;
    }
    return new;
}

static struct tagTimerInitiator
*InsertNewTimerInitiator(struct tagTimerInitiator **head)
{
    struct tagTimerInitiator *new =calloc(1, sizeof(struct tagTimerInitiator));

    if (new) {
        if (*head) {
            struct tagTimerInitiator *tmp = *head;
            while (tmp->next) tmp = tmp->next;
            tmp->next = new;
        }
        else *head = new;
    }
    return new;
}

static struct tagMacroData *InsertNewMacroData(struct tagMacroData **head)
{
    struct tagMacroData * new = calloc(1,sizeof(struct tagMacroData));

    if (new) {
        if (*head) {
            struct tagMacroData *tmp = *head;
            while (tmp->next) tmp = tmp->next;
            tmp->next = new;
        }
        else *head = new;
    }
    return new;
}

static struct tagEvent *InsertEvent(struct tagEvent *new,
                                    struct tagEvent **head)
{
    if (!new) {
        new = calloc(1,sizeof(struct tagEvent));
    }
    
    if (new) {
        new->next = NULL;
        
        if (*head) {
            struct tagEvent *tmp = *head;
            while (tmp->next) tmp = tmp->next;
            tmp->next = new;
        }
        else *head = new;
    }
    return new;
}


// Read and define macro data
// (until ENDM)
// If label is non-NULL, then we are defining a MACRODATA group
// and calls are not allowed
static struct tagMacroData * ReadMacroData(STREAM inStream, int forceDefine) 
{
    char token[32];
    struct tagMacroData *macPtr = NULL;
    int doesCall = 0;
    
    struct tagMacroData * CreateMacPtr(void) {
        if (!macPtr) {
            macPtr = InsertNewMacroData(&gMacroDataHead);
            if (!macPtr) {
                StreamError(inStream);
                AllocFailed();
                return NULL;
            }
            macPtr->referenceCount = (forceDefine ? 0 : 1);
        }
        return macPtr;
    }

    while (GetNextToken(inStream, token, sizeof(token))
           && strcasecmp(token, gEndM)) {
        // Main loop, expecting CALL, DELAY, EVENT, NAME
        if (DEBUG_MACRO)
            fprintf(stderr,"ReadMacroData token %s\n", token);

        if (!strcasecmp(token, gName)) {
            // NAME

            if (doesCall) {
                StreamError(inStream);
                fprintf(stderr,
                        "Attempt to add NAME to a CALLed macro\n");
                return NULL;
            }
            
            if (!CreateMacPtr()) {
                return NULL;
            }
    
            if (!GetNextToken(inStream, token, sizeof(token))) {
                // No token found
                UnexpectedEnd();
                return NULL;
            }
            if (DEBUG_MACRO)
                fprintf(stderr, " name token = %s\n", token);
            
            strncpy(macPtr->label, token, sizeof(macPtr->label)-1);
        } 
        else if (!strcasecmp(token, gDelay)) {
            // DELAY
            if (doesCall) {
                StreamError(inStream);
                fprintf(stderr,
                        "Attempt to add DELAY to a CALLed macro\n");
                return NULL;
            }
            
            if (!GetNextToken(inStream, token, sizeof(token))) {
                // No token found
                UnexpectedEnd();
                return NULL;
            }

            if (!CreateMacPtr()) return NULL;

            macPtr->delayMinutes = atoi(token);
            if (DEBUG_MACRO)
                fprintf(stderr," delay time = %d\n", macPtr->delayMinutes);
            
        } else if (!strcasecmp(token, gEvent)) {
            // EVENT
            struct tagEvent *newEvent;
            X10_UNIT_BITMAP exModMap = 0;
            X10_UNIT unit;
            
            if (doesCall) {
                StreamError(inStream);
                fprintf(stderr,
                        "Attempt to add EVENT to a CALLed macro\n");
                return NULL;
            }
            if (!CreateMacPtr()) return NULL;

            newEvent = ReadEvent(inStream);
            
            if (!newEvent) return NULL;

            if ((newEvent->function & 0x0f) == FN_DIM &&
                newEvent->dimAbsolute &&
                (exModMap = (newEvent->deviceMask &
                             gExtModules[newEvent->function >> 4])) != 0) {

                // Hey, we can replace some of the DIMTO's with
                // the nifty extended preset dim version
                if (DEBUG_MACRO)
                    fprintf(stderr, "Creating extended event(s)\n");

                for (unit = 0; unit < 16; unit++) {
                    if (exModMap & (1 << unit)) {
                        struct tagEvent * tmp;
                        
                        if ((tmp = calloc(1,sizeof(struct tagEvent))) == NULL) {
                            StreamError(inStream);
                            AllocFailed();
                            return NULL;
                        }
                        tmp->function = FN_EXT_CODE | (newEvent->function & 0xf0);
                        tmp->extended[0] = unit;
                        tmp->extended[1] = ((100-newEvent->dimLevel) * 62) / 100;
                        tmp->extended[2] = EFN_SET_BRIGHTNESS;
                        InsertEvent(tmp, &macPtr->firstEvent);
                    }
                }
                newEvent->deviceMask &= ~(exModMap);
                if (newEvent->deviceMask) {
                    // Some old-style devices left?
                    InsertEvent(newEvent, &macPtr->firstEvent);
                } else {
                    // Nope, we have it handled
                    free(newEvent);
                }
                
            } else InsertEvent(newEvent, &macPtr->firstEvent);
                
        } else if (!strcasecmp(token, gCall)) {
            // CALL
            struct tagMacroData *compare = gMacroDataHead;

            if (macPtr != NULL || forceDefine) {
                StreamError(inStream);
                fprintf(stderr,"Attempt to CALL inside a macro definition"
                        " or after another CALL\n");
                return NULL;
            }
            doesCall = 1;
            
            // We're calling another macro.  Find it!
            if (!GetNextToken(inStream, token, sizeof(token))) {
                // No token found
                UnexpectedEnd();
                return NULL;
            }
            
            // token contains the called macro name
            while (compare != NULL &&
                   strcasecmp(compare->label, token)) {
                if (DEBUG_MACRO)
                    fprintf(stderr, "Call compare: %s != %s\n",
                            compare->label, token);

                compare = compare->next;
            }
            if (compare == NULL) {
                // end of list.
                StreamError(inStream);
                fprintf(stderr, "Unknown macro %s\n", token);
                return NULL;
            }
            
            // Got it
            macPtr = compare;
            compare->referenceCount++;
            if (DEBUG_MACRO)
                fprintf(stderr," CALLED macro %s, new ref count %d\n", token,
                        compare->referenceCount);
            
        } else {
            StreamError(inStream);
            fprintf(stderr, "Unkown token %s\n", token);
        }
    } // While
    if (!macPtr) {
        fprintf(stderr, "Empty macro will be ignored\n");
    }
    else if (forceDefine && !macPtr->label[0]) {
        fprintf(stderr, "No NAME in MACRODATA; will be ignored\n");
    }
    
    return macPtr;
}

static char dayMaskStr[] = "SMTWTFSsmtwtfs";

static int ReadTimer(STREAM inStream) 
{
    char token[32];
    int i;
    // Zero it
    struct tagTimerInitiator * timer =
        InsertNewTimerInitiator(&gTimerInitiatorHead);

    if (DEBUG_MACRO)
        fprintf(stderr,"ReadTimer\n");

    if (timer == NULL) {
        StreamError(inStream);
        AllocFailed();
        return -1;
    }
    
    // This can be a little daunting because we have optional parameters

    // First, however, is positional and required
    // Day mask: SMTWTFS... Missing days are replaced by dash
    if (!GetNextToken(inStream, token, sizeof(token))
        || strlen(token) != 7) {
        StreamError(inStream);
        fprintf(stderr,"TIMER not followed by SMTWTFS mask.\n");
        return -1;
    }
    for (i = 0; i < 7 ; i++) {
        if (token[i] == dayMaskStr[i] || token[i] == dayMaskStr[i+7]) {
            timer->dayMask |= (1 << i);
        }
        else if (token[i] != '-') {
            StreamError(inStream);
            fprintf(stderr,"Date mask %s not compliant with SMTWTFS.\n", token);
            return -1;
        }
    }
    if (DEBUG_MACRO)
        fprintf(stderr," Day mask %s decoded as %x\n", token, timer->dayMask);

    while (GetNextToken(inStream, token, sizeof(token))
           && strcmp(token, gEndT)) {
        if (!strcmp(token, gTStart)) {
            if (timer->startMacroData) {
                StreamError(inStream);
                fprintf(stderr,"Multiple START sections!\n");
                return -1;
            }
            if ((timer->startMinutes = ReadTime(inStream)) < 0) {
                StreamError(inStream);
                fprintf(stderr,"Invalid START time\n");
                return -1;
            }
            if (DEBUG_MACRO)
                fprintf(stderr," Start at %d\n", timer->startMinutes);

            timer->startMacroData = ReadMacroData(inStream, 0);
            if (!timer->startMacroData) return -1;
            
            if (*timer->startMacroData->label == '\0') {
                strcpy(timer->startMacroData->label, "unlabeled start");
            }
        }
        else if (!strcmp(token, gTStop)) {
            if (timer->stopMacroData) {
                StreamError(inStream);
                fprintf(stderr,"Multiple STOP sections!\n");
                return -1;
            }
            if ((timer->stopMinutes = ReadTime(inStream)) < 0) {
                StreamError(inStream);
                fprintf(stderr,"Invalid STOP time\n");
                return -1;
            }
            
            if (DEBUG_MACRO)
                fprintf(stderr," Stop at %d\n", timer->stopMinutes);

            timer->stopMacroData = ReadMacroData(inStream, 0);
            if (!timer->stopMacroData) return -1;
            if (*timer->stopMacroData->label == '\0') {
                strcpy(timer->stopMacroData->label, "unlabeled stop");
            }
        }
        else if (!strcmp(token, gTBegin)) {
            if ((timer->beginDay = ReadDate(inStream)) < 0) {
                StreamError(inStream);
                fprintf(stderr,"Invalid BEGIN day.\n");
                return -1;
            }
            if (DEBUG_MACRO)
                fprintf(stderr," Begin at %d\n", timer->beginDay);
        }
        else if (!strcmp(token, gTEnd)) {
            if ((timer->endDay = ReadDate(inStream)) < 0) {
                StreamError(inStream);
                fprintf(stderr,"Invalid END day.\n");
                return -1;
            }
            if (DEBUG_MACRO)
                fprintf(stderr," End at %d\n", timer->endDay);
        }
    }
    // A little cleanup
    if (--timer->beginDay < 0) {
        timer->beginDay = 0;
    }
    
    if (--timer->endDay < 0) {
        timer->endDay = 365;
    }
    if (!timer->startMacroData) {
        StreamError(inStream);
        fprintf(stderr, "No START macro defined!\n");
        return -1;
    }
    if (!timer->stopMacroData) {
        // Force a stop time that will never execute
        timer->stopMinutes = -1;
    }
    return 0;
}

    
// Read entire file of macro defs
// 
// Top-level: expecting MACRODATA, MACRO, TIMER, or EXPORT lines
//
int ReadMacroFile(STREAM inStream) 
{
    char token[32];

    while (GetNextToken(inStream, token, sizeof(token))) {
        if (DEBUG_MACRO)
            fprintf(stderr,"ReadMacroFile: token %s\n", token);

        if (!strcasecmp(token, gKeyExtModule)) {
            X10_UNIT house;
            X10_UNIT_BITMAP devices;
            
            if (!GetNextToken(inStream, token, sizeof(token))) {
                UnexpectedEnd();
                return -1;
            }
            if (StrToUnits(token, &house, &devices) < 0) {
                StreamError(inStream);
                fprintf(stderr," invalid token for %s\n", gKeyExtModule);
                return -1;
            }
            // OR it in
            gExtModules[(house & 0xf0) >> 4] |= devices;
            
            
        } else if (!strcasecmp(token, gKeySymbol)) {
            if (symbolMax < MAX_SYMBOLS) {
                if (!GetNextToken(inStream, token, sizeof(token))) {
                    UnexpectedEnd();
                    return -1;
                }
                gSymbol[symbolMax].address = strtoul(token, NULL, 0);
                if (!GetNextToken(inStream,
                                  gSymbol[symbolMax].name,
                                  sizeof(gSymbol[symbolMax].name))) {
                    UnexpectedEnd();
                    return -1;
                }
                if (DEBUG_MACRO) {
                    fprintf(stderr, "Symbol %s at address %lx\n",
                            gSymbol[symbolMax].name,
                            gSymbol[symbolMax].address);
                }
                ++symbolMax;
            }
        } else if (!strcasecmp(token, gKeyAlias)) {
            char alias[32];
            char translation[256];
            
            if (!GetNextToken(inStream, alias, sizeof(alias))
                || !GetNextToken(inStream, translation, sizeof(translation))) {
                UnexpectedEnd();
                return -1;
            }
            if (DEBUG_MACRO)
                fprintf(stderr,"Adding alias %s->%s\n", alias, translation);

            AddAlias(alias, translation);

        } else if (!strcasecmp(token, gKeyMacroData)) {
            // Defining macrodata
            if (DEBUG_MACRO)
                fprintf(stderr,"ReadMacroFile: MACRODATA\n");

            if (!ReadMacroData(inStream, 1)) return -1;
        }
        else if (!strcasecmp(token, gKeyMacro)) {
            char onOff[8];
            struct tagMacroInitiator *new;
            
            if (!GetNextToken(inStream, onOff, sizeof(onOff))
                || !GetNextToken(inStream, token, sizeof(token))) {
                UnexpectedEnd();
                return -1;
            }
            new = InsertNewMacroInitiator(&gMacroInitiatorHead);
            if (!new) {
                StreamError(inStream);
                AllocFailed();
            }
            new->isOnFunction = !strcasecmp(onOff,gFnOn);
            new->initiatorUnit = StrToUnit(token);
            new->macroData = ReadMacroData(inStream, 0);
            if (!new->macroData) return -1;
            if (*new->macroData->label == '\0') {
                strcpy(new->macroData->label, "unlabeled macro");
            }

            if (DEBUG_MACRO)
                fprintf(stderr,"ReadMacroFile: MACRO %d,%x\n",
                        new->isOnFunction,
                        new->initiatorUnit);
        }
        else if (!strcasecmp(token, gKeyTimer)) {
            // Timer
            if (ReadTimer(inStream) < 0) {
                return -1;
            }
        }
        else if (!strcasecmp(token, gKeyExport)) {
            // Not implemented yet
            StreamError(inStream);
            fprintf(stderr,"Exports not implemented yet.\n");
            return -1;
        } else {
            // Error!
            StreamError(inStream);
            fprintf(stderr,"Unknown token %s\n", token);
            return -1;
        }
    } // Loop
    if (DEBUG_MACRO)
        fprintf(stderr,"End of Data\n");

    return 0;
}


// Linker phase

static int TimerInitiatorSize()
{
    // Return compiled size of timer initiators
    struct tagTimerInitiator *t = gTimerInitiatorHead;

    // There will always be the ff at the end
    int count = 1;

    while (t) {
        count += 9;
        t = t->next;
    }
    return count;
}

static int MacroInitiatorSize()
{
    struct tagMacroInitiator *m = gMacroInitiatorHead;
    int count = 2; // For trailing ffff
    
    while (m) {
        count += 3;
        m = m->next;
    }
    return count;
}

static int MacroDataSize()
{
    struct tagMacroData *m = gMacroDataHead;
    int count = 0;

    while (m) {
        // Skip over unused macros
        if (m->referenceCount) {
            // Timer offset + number of elements
            struct tagEvent *e = m->firstEvent;
            count += 2;
            
            while (e) {
                switch (e->function & 0x0f) {
                case FN_EXT_CODE:
                case FN_EXT_DATA_XFER:
                    count += 6;
                    break;
                case FN_DIM:
                case FN_BRIGHT:
                    count += 4;
                    break;
                default:
                    count += 3;
                }
                e = e->next;
            }
        }
        m = m->next;
    }
    return count;
}

static void LinkTimerInitiator(BYTE *s, int offset)
{
    // Return compiled size of timer initiators
    struct tagTimerInitiator *t = gTimerInitiatorHead;

    while (t) {
	unsigned startEpromAddress;
	unsigned stopEpromAddress;
	if (t->startMacroData)
	    startEpromAddress = t->startMacroData->epromAddress;
	else
	    startEpromAddress = 0;
	if (t->stopMacroData)
	    stopEpromAddress = t->stopMacroData->epromAddress;
	else
	    stopEpromAddress = 0;
        s[offset++] = t->dayMask;
        s[offset++] = t->beginDay & 0xff;
        s[offset++] = t->endDay & 0xff;
        s[offset++] = ((t->startMinutes /120) << 4) | (t->stopMinutes / 120);
        s[offset++] = t->startMinutes % 120 | ((t->beginDay & 0x100) ? 0x80 : 0);
        s[offset++] = t->stopMinutes % 120 | ((t->endDay & 0x100) ? 0x80 : 0);
        s[offset++] = (((startEpromAddress >> 8) & 0x0f) << 4) |
                       ((stopEpromAddress  >> 8) & 0x0f);
        s[offset++] = startEpromAddress & 0xff;
        s[offset++] = stopEpromAddress & 0xff;
        
        t = t->next;
    }
    s[offset++] = 0xff;
    
}

static void LinkMacroInitiator(BYTE *s, int offset)
{
    struct tagMacroInitiator *m = gMacroInitiatorHead;
    
    while (m) {
        s[offset++] = m->initiatorUnit;
        s[offset++] = (m->isOnFunction ? 0x80 : 0)
            | (m->macroData->epromAddress >> 8);
        s[offset++] = m->macroData->epromAddress & 0xff;
        m = m->next;
    }
    s[offset++] = 0xff;
    s[offset++] = 0xff;
}

static void LinkMacroData(BYTE *s, int offset)
{
    // Fill up buffer s starting at offset
    struct tagMacroData *m = gMacroDataHead;

    if (symbolMax > 0) {
        fprintf(stderr, "Resetting symbols\n");
    }
    symbolMax = 0;
    
    while (m) {
        // Skip over unused macros
        if (m->referenceCount) {
            // Timer offset + number of elements
            int countOffset;
            int count = 0;
            struct tagEvent *e = m->firstEvent;

            // Save eprom address
            m->epromAddress = offset;
            strncpy(gSymbol[symbolMax].name, m->label,
                     sizeof(gSymbol[symbolMax].name));
            gSymbol[symbolMax++].address = offset;
            
            if (DEBUG_MACRO)
                fprintf(stderr,"Macro %s at %x\n", m->label, offset);
            s[offset++] = (BYTE)m->delayMinutes;
            // Save event count
            countOffset = offset++;

            while (e) {
                if (DEBUG_MACRO)
                    fprintf(stderr,"Event at %x\n", offset);

                count++;
                // All events start with these three
                s[offset++] = (BYTE)e->function;
                s[offset++] = (BYTE)(e->deviceMask >> 8);
                s[offset++] = (BYTE)(e->deviceMask & 0xff);
            
                switch (e->function & 0x0f) {
                case FN_EXT_CODE:
                case FN_EXT_DATA_XFER:
                    s[offset++] = e->extended[0];
                    s[offset++] = e->extended[1];
                    s[offset++] = e->extended[2];
                    break;
                case FN_DIM:
                case FN_BRIGHT:
                    s[offset++] = ((e->dimLevel * 22) / 100) |
                        (e->dimAbsolute ? 0x80 : 0);
                    break;
                }
                e = e->next;
            }
            // Now we can pop in the event count
            s[countOffset] = count;
        }
        if (DEBUG_MACRO)
            fprintf(stderr,"Macro ends at %x\n", offset);

        m = m->next;
    }
}

void LookupSymbol(unsigned long address, char *buffer, int size)
{
    int i;

    for (i = 0; i < symbolMax; i++) {
        if (gSymbol[i].address == address) {
            strncpy(buffer, gSymbol[i].name, size);
            return;
        }
    }
    sprintf(buffer, "Unkown at %03lx", address);
}

void DumpSymbols(BYTE *s , int size, char * symbolFile)
{
    int i,j;
    FILE *outF = fopen(symbolFile, "w+");
    
    if (!outF) {
        return;
    }
    if (optVerbose) {
        fprintf(stderr, "Dumping symbols to %s\n",symbolFile);
    }
    
    fprintf(outF,
            "# CM11 Assembler Symbols\n"
            "# This file was generated automatically!\n"
            "# Aliases:\n");
    
    DumpAliases(outF);
    
    fprintf(outF,"\n# Symbols:\n");
    
    for (i = 0; i < symbolMax; i++) {
        fprintf(outF, "%s 0x%03lx \"%s\"\n", gKeySymbol, gSymbol[i].address,
                gSymbol[i].name);
    }
    
    fprintf(outF,"\n# Eprom image:\n");
    
    for (i = 0; i < size; i+= 0x10) {
        fprintf(outF,"\n# %03x:", i);
        for (j = 0; j < 0x10 && i+j < size; j++) 
            fprintf(outF," %02x", s[i+j]);
    }
    fprintf(outF,"\n");
    fclose(outF);
}

static BYTE * eprom;

int CreateEpromImage(BYTE ** retEprom)
{
    int timerOffset = 2;
    int macroOffset = timerOffset + TimerInitiatorSize();
    int macroDataOffset = macroOffset + MacroInitiatorSize();
    int totalSize = macroDataOffset + MacroDataSize();

    if (eprom) free(eprom);

    if (totalSize > 1024) {
        fprintf(stderr, "Eprom size %x is too large\n", totalSize);
        return -1;
    }

    // Allocate enough space for pad to nearest 16-bytes
    eprom = calloc(1, ((totalSize-1) & 0xff0) + 0x10);
    if (!eprom) {
        AllocFailed();
        return -1;
    }
    *retEprom = eprom;
    
    if (DEBUG_MACRO)
        fprintf(stderr," Image size = %x\n", totalSize);

    eprom[0] = macroOffset >> 8;
    eprom[1] = macroOffset & 0xff;

    // Must do the macro data first, since the link
    // performs actual address calculations used by the
    // initiator links
    LinkMacroData(eprom,macroDataOffset);
    LinkMacroInitiator(eprom,macroOffset);
    LinkTimerInitiator(eprom,timerOffset);

    return totalSize;
}

void CleanupMacroStuff()
{
    if (eprom) free(eprom);

    while (gMacroDataHead) {
        struct tagMacroData *tM = gMacroDataHead->next;
        
        while (gMacroDataHead->firstEvent) {
            struct tagEvent *tE = gMacroDataHead->firstEvent->next;
            free(gMacroDataHead->firstEvent);
            gMacroDataHead->firstEvent = tE;
        }
        free(gMacroDataHead);
        gMacroDataHead = tM;
    }

    while (gMacroInitiatorHead) {
        struct tagMacroInitiator *tMM = gMacroInitiatorHead->next;
        free(gMacroInitiatorHead);
        gMacroInitiatorHead = tMM;
    }
    
    while (gTimerInitiatorHead) {
        struct tagTimerInitiator *tTT = gTimerInitiatorHead->next;
        free(gTimerInitiatorHead);
        gTimerInitiatorHead = tTT;
    }
    symbolMax = 0;
    
}
