#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>

#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <malloc.h>

#include <termios.h>

#include "cm11.h"

#define TTY_BUFSIZE	128		/* size of one serial buffer	*/

static char		*in_buff,	/* line input/output buffers	*/
			*out_buff,
			*in_ptr,
			*out_ptr;

static int		in_size,	/* buffer sizes and counters	*/
			out_size,
			in_cnt,
			out_cnt,
			tty_fd = -1;	/* TTY file descriptor		*/

int InstallSerialDrivers(char * path)
{
    struct termios tty;
    int fd,i;
    int t;

    if ((fd = open(path, O_RDWR|O_NDELAY)) < 0) {
        fprintf(stderr, "tty: open(%s, RW): %s\n",
                path, strerror(errno));
        return -errno;
    }
#ifdef sun
    t = lockf(fd, F_TLOCK, 0);
#define LOCKNAME "lockf"
#else
    t = flock(fd, LOCK_EX | LOCK_NB);
#define LOCKNAME "flock"
#endif
    if (t < 0) {
        if (DEBUG_SERIAL)
            fprintf(stderr, "tty: %s(%s): %s\n", LOCKNAME,
                    path, strerror(errno));
        fprintf(stderr,"Serial device in use by another process.\n");
        
        return -errno;
    }
    
    tty_fd = fd;

    /* Size and allocate the I/O buffers. */
    in_size = TTY_BUFSIZE;
    out_size = in_size;
    in_buff = (char *) malloc(in_size+out_size);
  
    if (in_buff == NULL) {
	fprintf(stderr, "tty: open: cannot alloc(%d, %d) buffers (%d)\n",
                in_size, out_size, errno);
	return -errno;
    }
    out_buff = in_buff + in_size;
    in_cnt = 0;
    out_cnt = 0;
    in_ptr = in_buff;
    out_ptr = out_buff;
    out_size -= 4; /* safety */


    if (ioctl(tty_fd, TCGETS, &tty) < 0) {
        fprintf(stderr, "tty: get_state: %s\n", strerror(errno));
        return(-errno);
    }

    for(i = 0; i < NCCS; i++)
        tty.c_cc[i] = '\0';            /* no spec chr          */
    tty.c_cc[VMIN] = 1;
    tty.c_cc[VTIME] = 0;
    tty.c_iflag = (IGNBRK | IGNPAR);             /* input flags          */
    tty.c_oflag = (0);                           /* output flags         */
    tty.c_lflag = (0);                           /* local flags          */
    tty.c_cflag = (HUPCL|CREAD|CLOCAL);  /* UART flags           */
    tty.c_cflag &= ~CSTOPB ;
    tty.c_cflag |= CS8;
    tty.c_cflag &= ~(PARENB | PARODD);
    tty.c_cflag &= ~CBAUD;
    tty.c_cflag |= B4800;

    if (ioctl(tty_fd, TCSETS, &tty) < 0) {
        fprintf(stderr, "tty: set_state: %s\n", strerror(errno));
        return(-errno);
    }
    return 0;
}

int UninstallSerialDrivers(void)
{
    if (in_buff) free(in_buff);
    
    return close(tty_fd);
}

int DataAvailable()
{
    int s;

    if (in_cnt <= 0) {
	s = read(tty_fd, in_buff, in_size);

        if ((DEBUG_SERIAL) && s > 0) {
            int i;
            fprintf(stderr,"<<%d ",s);
            for (i = 0; i < s; i++) fprintf(stderr," %02x",(unsigned char)in_buff[i]);
            fprintf(stderr,"\n");
        }

	in_cnt = s;
	in_ptr = in_buff;
    }
    if (DEBUG_SERIAL)
        fprintf(stderr,"DA %d\n",in_cnt);
    
    return (in_cnt>0? in_cnt:0);
}


int DataAvailableWait(int timeoutSeconds)
{
    // select example stolen from pppd
    struct timeval timo;
    fd_set ready;
    int n;
    int da = DataAvailable();
    
    if (da) return da;
    
    FD_ZERO(&ready);
    FD_SET(tty_fd, &ready);

    timo.tv_sec  = timeoutSeconds;
    timo.tv_usec = 0;
    if (DEBUG_SERIAL)
        fprintf(stderr,"Waiting...");
    
    n = select(tty_fd + 1, &ready, NULL, &ready, timeoutSeconds ? &timo : NULL);
    if (DEBUG_SERIAL)
        fprintf(stderr,"...woke up %d!\n", n);
    
    if (n < 0 && errno != EINTR) {
	fprintf(stderr, "select: %d", errno);
        return 0;
    }
    return DataAvailable();
}


int DataQueued()
{
    return 0;
}

int GetSerialChar()
/* Gets character - no wait - returns -1 if not available now */
{
    int s;

    if (in_cnt <= 0) {
        DataAvailable();
    }

    if (in_cnt <= 0) {
        return -1;
    }

    s = (int) *in_ptr;
    s &= 0xFF;
    in_ptr++;
    in_cnt--;
    return(s);
}

unsigned char GetSerialCharWait()
{
    DataAvailableWait(0);
    return GetSerialChar();

}


int PutSerialCharBuffered(int c)
{
    int s;
    
    
    if ((out_cnt == out_size) || (c == -1)) {
        s = write(tty_fd, out_buff, out_cnt);
        if (s < 0 || s != out_cnt) {
            printf("write err %d\n",s);
            return(-1);
        }
        if (DEBUG_SERIAL) {

            fprintf(stderr, ">>%d ",out_cnt);
            for (s = 0; s < out_cnt; s++) fprintf(stderr," %02x",
                                                  (unsigned char)out_buff[s]);
            fprintf(stderr,"\n");
        }
        
        out_cnt = 0;
        out_ptr = out_buff;
    }
  
    if (c != -1) {
        *out_ptr = (char) c;
        out_ptr++;
        out_cnt++;
    }
    return(0);  
}

int PutSerialChar(BYTE c)
{
    PutSerialCharBuffered(c);
    return PutSerialCharBuffered(-1);
}


int WriteSerial(unsigned char *c,int len)
{
    int i;
    for (i=0; i<len; i++) PutSerialCharBuffered((int)*(c+i));
    PutSerialCharBuffered(-1);
    return 0;
}


int DataPurge()
{
    int numPurged = 0;
    
    while (GetSerialChar() > -1) {
        ++numPurged;
    }
    return numPurged;
}
